/*
 * playlistaspectsearchfilter.cpp
 *
 *  Created on: 18.04.2010
 *      Author: darkstar
 */

#include "playlistaspectsearchfilter.h"
#include "sqlite3.h"
#include "debug.h"
#include "compathack.h"

#define cGramSize 3
#define cGramMatchPercent 85

PlayListAspectSearchFilter::PlayListAspectSearchFilter()
	:	QObject(),
	 	PlayListAspectExtension(),
		filterText_(""),
		dbFilterText_(QString::null),
		nGramSearchEnabled_(true),
		nGramIndexUpdated_(false),
		stmtInsertNGramIndex_(NULL),
		stmtInsertTokenIndex_(NULL)
{

}

PlayListAspectSearchFilter::~PlayListAspectSearchFilter()
{

}

void PlayListAspectSearchFilter::attachedToAspect()
{
	connect(playListAspect()->playList()->mediaDatabase(), SIGNAL(preparingStatements(sqlite3*)),
			this, SLOT(mediaDatabasePreparingStatements(sqlite3*)));

	connect(playListAspect()->playList()->mediaDatabase(), SIGNAL(finalizingStatements(sqlite3*)),
			this, SLOT(mediaDatabaseFinalizingStatements(sqlite3*)));

	createSchema();
}

void PlayListAspectSearchFilter::removedFromAspect()
{
	disconnect(playListAspect()->playList()->mediaDatabase(), SIGNAL(finalizingStatements(sqlite3*)),
			this, SLOT(mediaDatabaseFinalizingStatements(sqlite3*)));

	disconnect(playListAspect()->playList()->mediaDatabase(), SIGNAL(preparingStatements(sqlite3*)),
			this, SLOT(mediaDatabasePreparingStatements(sqlite3*)));
}

void PlayListAspectSearchFilter::mediaDatabasePreparingStatements(sqlite3 *db)
{
	DENTERMETHOD;

	sqlite3_prepare_v2(db, "INSERT OR REPLACE INTO ngram_index VALUES(?1, ?2);", -1, &stmtInsertNGramIndex_, 0);

	sqlite3_prepare_v2(db, "INSERT OR REPLACE INTO token_index VALUES(?1, ?2, ?3);", -1, &stmtInsertTokenIndex_, 0);

	DEXITMETHOD;
}

void PlayListAspectSearchFilter::mediaDatabaseFinalizingStatements(sqlite3 *db)
{
	DENTERMETHOD;

	if (stmtInsertNGramIndex_)
	{
		sqlite3_finalize(stmtInsertNGramIndex_);
		stmtInsertNGramIndex_ = NULL;
	}

	if (stmtInsertTokenIndex_)
	{
		sqlite3_finalize(stmtInsertTokenIndex_);
		stmtInsertTokenIndex_ = NULL;
	}

	DEXITMETHOD;
}


void PlayListAspectSearchFilter::setFilterText(const QString &filter)
{
	if (filter != filterText_)
	{
		filterText_ = filter;
		update();
	}
}

const QString PlayListAspectSearchFilter::outputViewSource()
{
	return playListAspect()->prefix() + "textfilter";
}

void PlayListAspectSearchFilter::reset()
{
	dbFilterText_ = QString::null;
	filterText_ = "";
}

void PlayListAspectSearchFilter::execute()
{
	DENTERMETHOD;

	/*
	if (nGramSearchEnabled_ && !nGramIndexUpdated_)
	{
		recreateNGramIndex(3);
		nGramIndexUpdated_ = true;
	}
	*/

	QString inputView = inputViewSource();
	QString outputView = outputViewSource();

	if (dbFilterText_.isEmpty() || dbFilterText_ != filterText_)
	{
		if (nGramSearchEnabled_)
			nGramTextSearch(inputViewSource(), outputViewSource(), filterText_);
		else
			naiveTextSearch(inputViewSource(), outputViewSource(), filterText_);

		dbFilterText_ = filterText_;
	}

	DEXITMETHOD;
}

void PlayListAspectSearchFilter::naiveTextSearch(const QString &inputView, const QString &outputView, const QString &searchString)
{
	QString query =
		"DROP VIEW IF EXISTS " + outputView + "; "
		"CREATE TEMPORARY VIEW " + outputView + " AS "
		"SELECT " + inputView + ".playlist_id as playlist_id, " + inputView + ".media_id as media_id, idx, track, title, album, artist, genre, playtime, filesize "
		"FROM " + inputView + ", media "
		"WHERE " + inputView + ".media_id = media.media_id ";

	if (!searchString.isEmpty())
	{
		QString conditionString = "";
		QString safeLikeTokenString = "";

		// split input filter string into parts so we can build separate conditions for each...
		QStringList tokenList = tokenizeString(searchString, false, true);
		int tokenCount = tokenList.count();

		// build condition string for view...
		for (int i = 0; i < tokenCount; ++i)
		{
			safeLikeTokenString = dbSafeString("%" + tokenList[i] + "%");

			//conditionString += "(title LIKE '" + safeSearchString + "' OR artist LIKE '" + safeSearchString + "' OR album LIKE '" + safeSearchString + "')";

			// the following is faster but less accurate:
			conditionString += "(title || album || artist LIKE '" + safeLikeTokenString + "')";

			// only append AND if this is not the last search string
			if (i < tokenCount -1)
				conditionString += " AND ";
		}

		query += "AND (" + conditionString + ") ";
	}

	query += ";";

	DTIMERINIT(timer);
	sqlite3_exec(playListAspect()->playList()->mediaDatabase()->db(), query.toUtf8(), NULL, NULL, NULL);
	DOP(playListAspect()->playList()->mediaDatabase()->dbDebug(query));

	// the schema changed, so we need to finalize and re-prepare statements...
	playListAspect()->playList()->mediaDatabase()->ensurePreparedStatementsAreValid();

	DTIMERPRINT(timer, "query");
}

QStringList PlayListAspectSearchFilter::tokenizeString(const QString &input, bool simplify, bool reorder)
{
	QString inputString = simplify ? input.simplified() : input;

	// split input string into tokens...
#ifdef QT4
	QStringList tokenList = inputString.split(" ", QString::SkipEmptyParts);
#else
	QStringList tokenList = QStringList::split(" ", inputString, false);
#endif

	// Stupid but effective optimization:
	// Find the longest token and make it the first item in the list
	// in order to minimize the matches already in the most nested select.
	// SQLite does only provide the most efficient indices in the most nested select.
	if (reorder)
	{
		int tokenCount = tokenList.count();
		int longestTokenIndex = 0;

		for (int i = 0; i < tokenCount; ++i)
			if (tokenList[i].length() > tokenList[longestTokenIndex].length())
				longestTokenIndex = i;

		if (longestTokenIndex > 0)
			qSwap(tokenList[0], tokenList[longestTokenIndex]);
	}

	return tokenList;
}

void PlayListAspectSearchFilter::nGramTextSearch(const QString &inputView, const QString &outputView, const QString &searchString)
{
	QString query =
		"DROP TABLE IF EXISTS " + outputView + "; "
		"CREATE TEMPORARY TABLE " + outputView + " AS ";

	if (searchString.isEmpty())
	{
		query +=
			"SELECT " + inputView + ".playlist_id as playlist_id, " + inputView + ".media_id as media_id, idx, track, title, album, artist, genre, playtime, filesize "
			"FROM " + inputView + ", media "
			"WHERE " + inputView + ".media_id = media.media_id ";
	}
	else
	{
		QString token;
		QString source = "playlist";
		QString temp;
		QString nGramSet;
		QStringList nGrams;
		int nGramMatchMinCount;
		bool isLast;

		QStringList tokenList = tokenizeString(searchString, false, true);
		int tokenCount = tokenList.count();

		// build condition string for view...
		for (int i = 0; i < tokenCount; ++i)
		{
			isLast = i == tokenCount - 1;

			nGrams.clear();
			token = tokenList[i].lower();
			DPRINTF("token: %s", (const char*)token.toUtf8());
			createNGrams(token, nGrams, cGramSize, token.length() < cGramSize);
			nGramSet = "'" + nGrams.join("','") + "'";
			nGramMatchMinCount = qMax(1, (int)(((float)cGramMatchPercent / 100) * nGrams.count()));

			temp =
				"SELECT"
				" playlist.playlist_id AS playlist_id,"
				" playlist.media_id AS media_id,"
				" playlist.idx AS idx ";

			if (isLast)
				temp +=	", track, title, album, artist, genre, playtime, filesize ";

			temp += "FROM " + source + ",";

			if (isLast)
				temp += "media,";

			temp +=
				" token_index,"
				" ngram_index "
				"WHERE ";

			if (isLast)
				temp += " media.media_id = playlist.media_id AND";

			temp +=
				" token_index.media_id = playlist.media_id AND"
				" ngram_index.token_id = token_index.token_id AND"
				" ngram_index.ngram IN (" + nGramSet + ") "
				"GROUP BY"
				" token_index.media_id "
				"HAVING"
				" COUNT(ngram_index.token_id) >= " + QString::number(nGramMatchMinCount);

			if (!isLast)
				source = "(" + temp + ") as playlist";
		}

		query += temp;
	}

	query += ";";

	DTIMERINIT(timer);
	sqlite3_exec(playListAspect()->playList()->mediaDatabase()->db(), query.toUtf8(), NULL, NULL, NULL);
	DOP(playListAspect()->playList()->mediaDatabase()->dbDebug(query));

	// the schema changed, so we need to finalize and re-prepare statements...
	playListAspect()->playList()->mediaDatabase()->ensurePreparedStatementsAreValid();

	DTIMERPRINT(timer, "query");
}

void PlayListAspectSearchFilter::createNGrams(const QString &token, QStringList &outputList, int gramSize, bool padded)
{
	QString input;
	if (padded)
	{
		input = token.rightJustify(token.length() + gramSize - 1, '%');
		input = input.leftJustify(input.length() + gramSize - 1, '%');
	}
	else
	{
		input = token;
		gramSize = qMin(gramSize, input.length());
	}

	//DPRINTF("input: %s", (const char *)input.toUtf8());

	for (int i = 0; i < input.length() - gramSize + 1; ++i)
		outputList.append(input.mid(i, gramSize));
}

void PlayListAspectSearchFilter::tokenizeGrammifyAndInsert(const QString &input, const unsigned long mediaID, const int gramSize)
{
	QStringList tokenList = tokenizeString(input, true, false);
	QStringList gramList;

    for (QStringList::Iterator it = tokenList.begin(); it != tokenList.end(); ++it)
    {
		sqlite3_bind_null(stmtInsertTokenIndex_, 1);
		sqlite3_bind_text(stmtInsertTokenIndex_, 2, (*it).toUtf8(), -1, SQLITE_TRANSIENT);
    	sqlite3_bind_int(stmtInsertTokenIndex_, 3, mediaID);
		sqlite3_step(stmtInsertTokenIndex_);
		unsigned long tokenID = sqlite3_last_insert_rowid(playListAspect()->playList()->mediaDatabase()->db());
		sqlite3_reset(stmtInsertTokenIndex_);

    	createNGrams((*it).lower(), gramList, gramSize, true);
		for (QStringList::Iterator it2 = gramList.begin(); it2 != gramList.end(); ++it2)
		{
			sqlite3_bind_text(stmtInsertNGramIndex_, 1, (*it2).toUtf8(), -1, SQLITE_TRANSIENT);
			sqlite3_bind_int(stmtInsertNGramIndex_, 2, tokenID);
			sqlite3_step(stmtInsertNGramIndex_);
			sqlite3_reset(stmtInsertNGramIndex_);
		}

		gramList.clear();
    }
}

void PlayListAspectSearchFilter::recreateNGramIndex(const int gramSize)
{
	DTIMERINIT(timer);

	MediaDatabase *mediaDatabase = playListAspect()->playList()->mediaDatabase();

	mediaDatabase->beginUpdate();

	sqlite3_exec(
		mediaDatabase->db(),
		"DROP TABLE IF EXISTS ngram_index;\n"
		"DROP TABLE IF EXISTS token_index;\n"
		,
		NULL, NULL, NULL
	);
	createSchema();

	sqlite3_stmt *vm;
	sqlite3_prepare_v2(mediaDatabase->db(), "SELECT media_id, title, album, artist FROM media;", -1, &vm, 0);

	unsigned long mediaID;
	QString title;
	QString album;
	QString artist;

	while (sqlite3_step(vm) == SQLITE_ROW)
	{
		mediaID = sqlite3_column_int(vm, 0);
		title = QString::fromUtf8((const char *)sqlite3_column_text(vm, 1));
		album = QString::fromUtf8((const char *)sqlite3_column_text(vm, 2));
		artist = QString::fromUtf8((const char *)sqlite3_column_text(vm, 3));

		tokenizeGrammifyAndInsert(title, mediaID, gramSize);
		tokenizeGrammifyAndInsert(album, mediaID, gramSize);
		tokenizeGrammifyAndInsert(artist, mediaID, gramSize);
	}

	sqlite3_finalize(vm);

	mediaDatabase->endUpdate();
	mediaDatabase->ensurePreparedStatementsAreValid();

	DTIMERPRINT(timer, "updateNGramIndex");
}

void PlayListAspectSearchFilter::createSchema()
{
	MediaDatabase *mediaDatabase = playListAspect()->playList()->mediaDatabase();

	mediaDatabase->beginUpdate();

	sqlite3_exec(
		mediaDatabase->db(),
		"CREATE TABLE IF NOT EXISTS ngram_index (\n"
		"  ngram TEXT,\n"
		"  token_id INTEGER\n"
	    ");\n"

		"CREATE TABLE IF NOT EXISTS token_index (\n"
		"  token_id INTEGER PRIMARY KEY,\n"
		"  token TEXT,\n"
		"  media_id INTEGER\n"
	    ");\n"

		"CREATE INDEX idx_ngram_index_ngram ON ngram_index (ngram);\n"
		"CREATE INDEX idx_ngram_index_token_id ON ngram_index (token_id);\n"
		"CREATE INDEX idx_token_index_media_id ON token_index (media_id);\n"
		,
		NULL, NULL, NULL
	);

	sqlite3_create_function(mediaDatabase->db(), "addNGramIndexFor", 4, SQLITE_UTF8, this, &PlayListAspectSearchFilter::sqlite3_addNGramIndexFor, NULL, NULL);

	sqlite3_exec(
		mediaDatabase->db(),
		"CREATE TRIGGER IF NOT EXISTS trg_delete_media_token_index DELETE ON media\n"
		"BEGIN\n"
		"	DELETE FROM token_index WHERE media_id = OLD.media_id;\n"
		"END;\n"

		"CREATE TRIGGER IF NOT EXISTS trg_delete_token_ngram_index DELETE ON token_index\n"
		"BEGIN\n"
		"	DELETE FROM ngram_index WHERE token_id = OLD.token_id;\n"
		"END;\n"

		"CREATE TRIGGER IF NOT EXISTS trg_insert_media_ngram_index INSERT ON media\n"
		"BEGIN\n"
		"	DELETE FROM token_index WHERE media_id = NEW.media_id;\n"
		"	SELECT addNGramIndexFor(NEW.media_id, NEW.title, NEW.album, NEW.artist);\n"
		"END;\n"

		"CREATE TRIGGER IF NOT EXISTS trg_update_media_ngram_index UPDATE OF title, album, artist ON media\n"
		"BEGIN\n"
		"	DELETE FROM token_index WHERE media_id = NEW.media_id;\n"
		"	SELECT addNGramIndexFor(NEW.media_id, NEW.title, NEW.album, NEW.artist);\n"
		"END;\n"
		,
		NULL, NULL, NULL
	);

	mediaDatabase->endUpdate();
	mediaDatabase->ensurePreparedStatementsAreValid();
}

void PlayListAspectSearchFilter::sqlite3_addNGramIndexFor(sqlite3_context *context, int, sqlite3_value **params)
{
	DENTERMETHOD;

	if (sqlite3_value_type(params[0]) == SQLITE_INTEGER &&
		sqlite3_value_type(params[1]) == SQLITE_TEXT &&
		sqlite3_value_type(params[2]) == SQLITE_TEXT &&
		sqlite3_value_type(params[3]) == SQLITE_TEXT)
	{
		PlayListAspectSearchFilter *aspectSource = static_cast<PlayListAspectSearchFilter *>(sqlite3_user_data(context));

		unsigned long mediaID = sqlite3_value_int(params[0]);
		QString title = QString::fromUtf8((const char *)sqlite3_value_text(params[1]));
		QString album = QString::fromUtf8((const char *)sqlite3_value_text(params[2]));
		QString artist = QString::fromUtf8((const char *)sqlite3_value_text(params[3]));

		DPRINTF("mediaID: %d", mediaID);
		DPRINTF("title: %s", (const char *)title.toUtf8());
		DPRINTF("album: %s", (const char *)album.toUtf8());
		DPRINTF("artist: %s", (const char *)artist.toUtf8());

		aspectSource->tokenizeGrammifyAndInsert(title, mediaID, cGramSize);
		aspectSource->tokenizeGrammifyAndInsert(album, mediaID, cGramSize);
		aspectSource->tokenizeGrammifyAndInsert(artist, mediaID, cGramSize);
	}

	DEXITMETHOD;
}

void PlayListAspectSearchFilter::saveExtensionState(QStringList &dst)
{
	dst.clear();
	dst.append(filterText_);
}

void PlayListAspectSearchFilter::loadExtensionState(QStringList &src)
{
	if (src.count() == 1)
		setFilterText(src[0]);
}
